__all__ = ["DualAttentionModel", "InputAttentionModel"]

import os

import numpy as np
import matplotlib.pyplot as plt
from easy_mpl import imshow

from .backend import tf
from .backend import keras
from .functional import Model as FModel
from ai4water.utils.utils import print_something
from ai4water.nn_tools import check_act_fn
from ai4water.preprocessing import DataSet
from ai4water.models.tensorflow.layer_definition import MyTranspose, MyDot
from ai4water.utils.utils import plot_activations_along_inputs

layers = keras.layers
KModel = keras.models.Model


class DALSTM(keras.layers.Layer):

    def __init__(
            self,
            enc_config: dict = None,
            dec_config: dict = None,
            drop_remainder: bool = True,
            teacher_forcing: bool = False,
            **kwargs
    ):
        self.enc_config = enc_config
        self.dec_config = dec_config
        self.drop_remainder = drop_remainder
        self.teacher_forcing = teacher_forcing

        super().__init__(**kwargs)
        raise NotImplementedError


class DualAttentionModel(FModel):
    """
    This is Dual-Attention LSTM model of Qin_ et al., 2017. The code is adopted
    from this_ repository

    Example:
        >>> from ai4water import DualAttentionModel
        >>> from ai4water.datasets import busan_beach
        >>> data = busan_beach()
        >>> model = DualAttentionModel(lookback=5,
        ...                            input_features=data.columns.tolist()[0:-1],
        ...                            output_features=data.columns.tolist()[-1:])
        If you do not wish to feed previous output as input to the model, you
        can set teacher forcing to False. The drop_remainder argument must be
        set to True in such a case.
        >>> model = DualAttentionModel(teacher_forcing=False, batch_size=4,
        ...                            drop_remainder=True, ts_args={'lookback':5})
        >>> model.fit(data=data)

    .. _Qin:
        https://arxiv.org/abs/1704.02971

    .. _this:
        https://github.com/chensvm/A-Dual-Stage-Attention-Based-Recurrent-Neural-Network-for-Time-Series-Prediction
    """
    _enc_config = {'n_h': 20,  # length of hidden state m
                   'n_s': 20,  # length of hidden state m
                   'm': 20,  # length of hidden state m
                   'enc_lstm1_act': None,
                   'enc_lstm2_act': None,
                                }
    # arguments for decoder/outputAttention in Dual stage attention
    _dec_config = {
        'p': 30,
        'n_hde0': 30,
        'n_sde0': 30
    }

    def __init__(
            self,
            enc_config: dict = None,
            dec_config: dict = None,
            teacher_forcing: bool = True,
            **kwargs
    ):
        """

        Arguments:
            enc_config:
                dictionary defining configuration of encoder/input attention. It must
                have following three keys

                    - n_h: 20
                    - n_s: 20
                    - m: 20
                    - enc_lstm1_act: None
                    - enc_lstm2_act: None

            dec_config:
                dictionary defining configuration of decoder/output attention. It must have
                following three keys

                    - p: 30
                    - n_hde0: None
                    - n_sde0: None

            teacher_forcing:
                Whether to use the prvious target/observation as input or not. If
                yes, then the model will require 2 inputs. The first input will be
                of shape (num_examples, lookback, num_inputs) while the second input
                will be of shape (num_examples, lookback-1, 1). This second input is
                supposed to be the target variable observed at previous time step.

            kwargs :
                The keyword arguments for the [ai4water's Model][ai4water.Model] class
        """
        self.method = 'dual_attention'
        if enc_config is None:
            enc_config = DualAttentionModel._enc_config
        else:
            assert isinstance(enc_config, dict)

        if dec_config is None:
            dec_config = DualAttentionModel._dec_config
        else:
            assert isinstance(dec_config, dict)
        self.enc_config = enc_config
        self.dec_config = dec_config

        super(DualAttentionModel, self).__init__(teacher_forcing=teacher_forcing, **kwargs)

        setattr(self, 'category', "DL")

    @property
    def mode(self) ->str:
        self.config['mode'] = "regression"
        return self.config['mode']

    def build(self, input_shape=None):

        self.config['dec_config'] = self.dec_config
        self.config['enc_config'] = self.enc_config
        setattr(self, 'batch_size', self.config['batch_size'])
        setattr(self, 'drop_remainder', self.config['drop_remainder'])

        self.de_LSTM_cell = layers.LSTM(self.dec_config['p'], return_state=True, name='decoder_LSTM')
        self.de_densor_We = layers.Dense(self.enc_config['m'])

        if self.config['drop_remainder']:
            h_de0 = tf.zeros((self.batch_size, self.dec_config['n_hde0']), name='dec_1st_hidden_state')
            s_de0 = tf.zeros((self.batch_size, self.dec_config['n_sde0']), name='dec_1st_cell_state')
        else:
            h_de0 = layers.Input(shape=(self.dec_config['n_hde0'],), name='dec_1st_hidden_state')
            s_de0 = layers.Input(shape=(self.dec_config['n_sde0'],), name='dec_1st_cell_state')

        input_y = None
        if self.teacher_forcing and self.drop_remainder:
            input_y = layers.Input(batch_shape=(self.batch_size, self.lookback - 1, self.num_outs), name='input_y')
        elif not self.drop_remainder:
            input_y = layers.Input(shape=(self.lookback - 1, self.num_outs), name='input_y')

        if self.drop_remainder:
            enc_input = keras.layers.Input(batch_shape=(self.batch_size, self.lookback, self.num_ins), name='enc_input')
        else:
            enc_input = keras.layers.Input(shape=(self.lookback, self.num_ins), name='enc_input')

        enc_lstm_out, s0, h0 = self._encoder(enc_input, self.config['enc_config'])

        # originally the last dimentions was -1 but I put it equal to 'm'
        # eq 11 in paper
        enc_out = layers.Reshape((self.lookback, self.enc_config['m']), name='enc_out_eq_11')(enc_lstm_out)

        h, context = self.decoder_attention(enc_out, input_y, s_de0, h_de0)
        h = layers.Reshape((self.num_outs, self.dec_config['p']))(h)
        # concatenation of decoder hidden state and the context vector.
        last_concat = layers.Concatenate(axis=2, name='last_concat')([h, context])  # (None, 1, 50)

        # original it was not defined but in tf-keras we need to define it
        sec_dim = self.enc_config['m'] + self.dec_config['p']
        last_reshape = layers.Reshape((sec_dim,), name='last_reshape')(last_concat)  # (None, 50)

        result = layers.Dense(self.dec_config['p'], name='eq_22')(last_reshape)  # (None, 30)  # equation 22

        output = layers.Dense(self.num_outs)(result)
        if self.forecast_len>1:
            output = layers.Reshape(target_shape=(self.num_outs, self.forecast_len))(output)

        initial_input = [enc_input]

        if input_y is not None:
            initial_input.append(input_y)

        if self.config['drop_remainder']:
            self._model = self.compile(model_inputs=initial_input, outputs=output)
        else:
            self._model = self.compile(model_inputs=initial_input + [s0, h0, s_de0, h_de0], outputs=output)

        return

    def _encoder(self, enc_inputs, config, lstm2_seq=True, suf: str = '1', s0=None, h0=None, num_ins=None):

        if num_ins is None:
            num_ins = self.num_ins

        self.en_densor_We = layers.Dense(self.lookback, name='enc_We_'+suf)
        _config, act_str = check_act_fn({'activation': config['enc_lstm1_act']})
        self.en_LSTM_cell = layers.LSTM(config['n_h'], return_state=True, activation=_config['activation'],
                                        name='encoder_LSTM_'+suf)
        config['enc_lstm1_act'] = act_str

        # initialize the first cell state
        if s0 is None:
            if self.drop_remainder:
                s0 = tf.zeros((self.batch_size, config['n_s']), name=f'enc_first_cell_state_{suf}')
            else:
                s0 = layers.Input(shape=(config['n_s'],), name='enc_first_cell_state_' + suf)
        # initialize the first hidden state
        if h0 is None:
            if self.drop_remainder:
                h0 = tf.zeros((self.batch_size, config['n_h']), name=f'enc_first_hidden_state_{suf}')
            else:
                h0 = layers.Input(shape=(config['n_h'],), name='enc_first_hidden_state_' + suf)

        enc_attn_out = self.encoder_attention(enc_inputs, s0, h0, num_ins, suf)

        enc_lstm_in = layers.Reshape((self.lookback, num_ins), name='enc_lstm_input_'+suf)(enc_attn_out)

        _config, act_str = check_act_fn({'activation': config['enc_lstm2_act']})
        enc_lstm_out = layers.LSTM(config['m'], return_sequences=lstm2_seq, activation=_config['activation'],
                                   name='LSTM_after_encoder_'+suf)(enc_lstm_in)  # h_en_all
        config['enc_lstm2_act'] = act_str

        return enc_lstm_out, h0, s0

    def one_encoder_attention_step(self, h_prev, s_prev, x, t, suf: str = '1'):
        """
        :param h_prev: previous hidden state
        :param s_prev: previous cell state
        :param x: (T,n),n is length of input series at time t,T is length of time series
        :param t: time-step
        :param suf: str, Suffix to be attached to names
        :return: x_t's attention weights,total n numbers,sum these are 1
        """
        _concat = layers.Concatenate()([h_prev, s_prev])  # (none,1,2m)
        result1 = self.en_densor_We(_concat)   # (none,1,T)
        result1 = layers.RepeatVector(x.shape[2],)(result1)  # (none,n,T)
        x_temp = MyTranspose(axis=(0, 2, 1))(x)  # X_temp(None,n,T)
        # (none,n,T) Ue(T,T), Ue * Xk in eq 8 of paper
        result2 = MyDot(self.lookback, name='eq_8_mul_'+str(t)+'_'+suf)(x_temp)
        result3 = layers.Add()([result1, result2])  # (none,n,T)
        result4 = layers.Activation(activation='tanh')(result3)  # (none,n,T)
        result5 = MyDot(1)(result4)
        result5 = MyTranspose(axis=(0, 2, 1), name='eq_8_' + str(t)+'_'+suf)(result5)  # etk/ equation 8
        alphas = layers.Activation(activation='softmax', name='eq_9_'+str(t)+'_'+suf)(result5)  # equation 9

        return alphas

    def encoder_attention(self, _input, _s0, _h0, num_ins, suf: str = '1'):

        s = _s0
        _h = _h0
        # initialize empty list of outputs
        attention_weight_t = None
        for t in range(self.lookback):

            _context = self.one_encoder_attention_step(_h, s, _input, t, suf=suf)  # (none,1,n)
            x = layers.Lambda(lambda x: _input[:, t, :])(_input)
            x = layers.Reshape((1, num_ins))(x)

            _h, _, s = self.en_LSTM_cell(x, initial_state=[_h, s])
            if t != 0:
                # attention_weight_t = layers.Merge(mode='concat', concat_axis=1,
                #                                    name='attn_weight_'+str(t))([attention_weight_t,
                #                                                                _context])
                attention_weight_t = layers.Concatenate(
                    axis=1,
                    name='attn_weight_'+str(t)+'_'+suf)([attention_weight_t, _context])
            else:
                attention_weight_t = _context

        # get the driving input series
        enc_output = layers.Multiply(name='enc_output_'+suf)([attention_weight_t, _input])  # equation 10 in paper

        return enc_output

    def one_decoder_attention_step(self, _h_de_prev, _s_de_prev, _h_en_all, t):
        """
        :param _h_de_prev: previous hidden state
        :param _s_de_prev: previous cell state
        :param _h_en_all: (None,T,m),n is length of input series at time t,T is length of time series
        :param t: int, timestep
        :return: x_t's attention weights,total n numbers,sum these are 1
        """
        # concatenation of the previous hidden state and cell state of the LSTM unit in eq 12
        _concat = layers.Concatenate(name='eq_12_'+str(t))([_h_de_prev, _s_de_prev])  # (None,1,2p)
        result1 = self.de_densor_We(_concat)   # (None,1,m)
        result1 = layers.RepeatVector(self.lookback)(result1)  # (None,T,m)
        result2 = MyDot(self.enc_config['m'])(_h_en_all)

        result3 = layers.Add()([result1, result2])  # (None,T,m)
        result4 = layers.Activation(activation='tanh')(result3)  # (None,T,m)
        result5 = MyDot(1)(result4)

        beta = layers.Activation(activation='softmax', name='eq_13_'+str(t))(result5)    # equation 13
        _context = layers.Dot(axes=1, name='eq_14_'+str(t))([beta, _h_en_all])  # (1,m)  # equation 14 in paper
        return _context

    def decoder_attention(self, _h_en_all, _y, _s0, _h0):
        s = _s0
        _h = _h0

        for t in range(self.lookback-1):
            _context = self.one_decoder_attention_step(_h, s, _h_en_all, t)  # (batch_size, 1, 20)

            # if we want to use the true value of target of previous timestep as input then we will use _y
            if self.teacher_forcing:
                y_prev = layers.Lambda(lambda y_prev: _y[:, t, :])(_y)  # (batch_size, lookback, 1) -> (batch_size, 1)
                y_prev = layers.Reshape((1, self.num_outs))(y_prev)   # -> (batch_size, 1, 1)

                # concatenation of decoder input and computed context vector  # ??
                y_prev = layers.Concatenate(axis=2)([y_prev, _context])   # (None,1,21)
            else:
                y_prev = _context

            y_prev = layers.Dense(self.num_outs, name='eq_15_'+str(t))(y_prev)  # (None,1,1),  Eq 15  in paper

            _h, _, s = self.de_LSTM_cell(y_prev, initial_state=[_h, s])   # eq 16  ??

        _context = self.one_decoder_attention_step(_h, s, _h_en_all, 'final')
        return _h, _context

    def fetch_data(self, x, y, source, **kwargs):

        if self.teacher_forcing:
            x, prev_y, labels = getattr(self.dh_, f'{source}_data')(**kwargs)
        else:
            x, labels = getattr(self.dh_, f'{source}_data')(**kwargs)
            prev_y = None

        n_s_feature_dim = self.enc_config['n_s']
        n_h_feature_dim = self.enc_config['n_h']
        p_feature_dim = self.dec_config['p']

        if kwargs.get('use_datetime_index', False):  # during deindexification, first feature will be removed.
            n_s_feature_dim += 1
            n_h_feature_dim += 1
            p_feature_dim += 1
            idx = np.expand_dims(x[:, 1:, 0], axis=-1)   # extract the index from x

            if self.use_true_prev_y:
                prev_y = np.concatenate([prev_y, idx], axis=2)  # insert index in prev_y

        other_inputs = []
        if not self.drop_remainder:
            s0 = np.zeros((x.shape[0], n_s_feature_dim))
            h0 = np.zeros((x.shape[0], n_h_feature_dim))

            h_de0 = s_de0 = np.zeros((x.shape[0], p_feature_dim))
            other_inputs = [s0, h0, s_de0, h_de0]

        if self.teacher_forcing:
            return [x, prev_y] + other_inputs, labels
        else:
            return [x] + other_inputs, labels

    def training_data(self, x=None, y=None, data='training', key=None):

        if x is not None:
            return x, y

        self._maybe_dh_not_set(data=data)
        return self.fetch_data(x=x, y=y, source='training', key=key)

    def validation_data(self, x=None, y=None, data='validation', **kwargs):
        self._maybe_dh_not_set(data=data)
        return self.fetch_data(x=x, y=y, source='validation', **kwargs)

    def test_data(self, x=None, y=None, data='test',  **kwargs):
        self._maybe_dh_not_set(data=data)
        return self.fetch_data(x=x, y=y, source='test', **kwargs)

    def _maybe_dh_not_set(self, data):
        """if dh_ has not been set yet, try to create it using data argument if
        possible"""
        if isinstance(data, str) and data not in ['training', 'test', 'validation']:
            self.dh_ = DataSet(data=data, **self.data_config)
        elif not isinstance(data, str):
            self.dh_ = DataSet(data=data, **self.data_config)
        return

    def interpret(self, data='training', **kwargs):
        import matplotlib
        matplotlib.rcParams.update(matplotlib.rcParamsDefault)

        self.plot_act_along_inputs(f'attn_weight_{self.lookback - 1}_1',
                                   data=data,
                                   **kwargs)
        return

    def get_attention_weights(
            self,
            layer_name: str=None,
            data = 'training',
    )->np.ndarray:
        """
        Parameters
        ----------
            layer_name : str, optional
                the name of attention layer. If not given, the final attention
                layer will be used.
            data : str, optional
                the data to make forward pass to get attention weghts. Possible
                values are

                - ``training``
                - ``validation``
                - ``test``

        Returns
        -------
            a numpy array of shape (num_examples, lookback, num_ins)
        """
        layer_name = layer_name or f'attn_weight_{self.lookback - 1}_1'

        assert isinstance(layer_name, str), f"""
            layer_name must be a string, not of {layer_name.__class__.__name__} type
            """

        from ai4water.postprocessing.visualize import Visualize

        kwargs = {}
        if self.config['drop_remainder']:
            kwargs['batch_size'] = self.config['batch_size']
        activation = Visualize(model=self).get_activations(layer_names=layer_name, data=data, **kwargs)

        activation = activation[layer_name]  # (num_examples, lookback, num_ins)

        return activation

    def plot_act_along_inputs(
            self,
            layer_name: str,
            name: str = None,
            vmin=None,
            vmax=None,
            data='training',
            show=False
    ):

        if not os.path.exists(self.act_path):
            os.makedirs(self.act_path)

        data_name = name or data

        activation = self.get_attention_weights(layer_name=layer_name, data=data)

        act_avg_over_examples = np.mean(activation, axis=0)  # (lookback, num_ins)

        lookback = self.config['ts_args']['lookback']
        data, _ = getattr(self, f'{data}_data')()

        predictions, observations = self.predict(process_results=False, data=data, return_true=True)

        plt.close('all')

        fig, axis = plt.subplots()

        ytick_labels = [f"t-{int(i)}" for i in np.linspace(lookback - 1, 0, lookback)]
        _, im = imshow(act_avg_over_examples,
                       ax=axis,
                       aspect="auto",
                       yticklabels=ytick_labels,
                       ylabel='lookback steps',
                       show=False
                       )

        axis.set_xticks(np.arange(self.num_ins))
        axis.set_xticklabels(self.input_features, rotation=90)
        fig.colorbar(im, orientation='horizontal', pad=0.3)
        plt.savefig(os.path.join(self.act_path, f'acts_avg_over_examples_{data_name}'),
                    dpi=400, bbox_inches='tight')
        plt.close('all')

        data = self.inputs_for_attention(data)

        plot_activations_along_inputs(
            data=data,
            activations=activation,
            observations=observations,
            predictions=predictions,
            in_cols=self.input_features,
            out_cols=self.output_features,
            lookback=lookback,
            name=data_name,
            path=self.act_path,
            vmin=vmin,
            vmax=vmax,
            show=show
        )
        return

    def plot_act_along_lookback(self, activations, sample=0):

        assert isinstance(activations, np.ndarray)

        activation = activations[sample, :, :]
        act_t = activation.transpose()

        fig, axis = plt.subplots()

        for idx, _name in enumerate(self.input_features):
            axis.plot(act_t[idx, :], label=_name)

        axis.set_xlabel('Lookback')
        axis.set_ylabel('Input attention weight')
        axis.legend(loc="best")

        plt.show()
        return

    def inputs_for_attention(self, inputs):
        """ returns the inputs for attention mechanism """
        if isinstance(inputs, list):
            inputs = inputs[0]

        inputs = inputs[:, -1, :]  # why 0, why not -1

        assert inputs.shape[1] == self.num_ins

        return inputs

    def _transform_x(self, x, name):
        """transforms x and puts the transformer in config witht he key name"""
        feature_names = [
            self.input_features,
            [f"{i}" for i in range(self.enc_config['n_s'])],
            [f"{i}" for i in range(self.enc_config['n_h'])],
            [f"{i}" for i in range(self.dec_config['n_hde0'])],
            [f"{i}" for i in range(self.dec_config['n_sde0'])],
        ]
        transformation = [self.config['x_transformation'], None, None, None, None]
        if self.teacher_forcing:
            feature_names.insert(1, self.output_features)
            transformation.insert(1, self.config['y_transformation'])
        return self._transform(x, name, transformation, feature_names)


class InputAttentionModel(DualAttentionModel):
    """
    InputAttentionModel is same as DualAttentionModel with output attention/decoder part
    removed.
    Example:
        >>> from ai4water import InputAttentionModel
        >>> from ai4water.datasets import busan_beach
        >>> model = InputAttentionModel(
        ... input_features=busan_beach().columns.tolist()[0:-1],
        ... output_features=busan_beach().columns.tolist()[-1:])
        >>> model.fit(data=busan_beach())
    """
    def __init__(self, *args, teacher_forcing=False, **kwargs):
        super(InputAttentionModel, self).__init__(*args, teacher_forcing=teacher_forcing, **kwargs)

    def build(self, input_shape=None):

        self.config['enc_config'] = self.enc_config
        setattr(self, 'batch_size', self.config['batch_size'])
        setattr(self, 'drop_remainder', self.config['drop_remainder'])

        setattr(self, 'method', 'input_attention')
        print('building input attention model')

        enc_input = keras.layers.Input(shape=(self.lookback, self.num_ins), name='enc_input1')
        lstm_out, h0, s0 = self._encoder(enc_input, self.enc_config, lstm2_seq=False)

        act_out = layers.LeakyReLU()(lstm_out)
        predictions = layers.Dense(self.num_outs)(act_out)
        if self.forecast_len>1:
            predictions = layers.Reshape(target_shape=(self.num_outs, self.forecast_len))(predictions)
        if self.verbosity > 2:
            print('predictions: ', predictions)

        inputs = [enc_input]
        if not self.drop_remainder:
            inputs = inputs + [s0, h0]
        self._model = self.compile(model_inputs=inputs, outputs=predictions)

        return

    def fetch_data(self, x, y, source, **kwargs):

        data = getattr(self.dh_, f'{source}_data')(**kwargs)
        if self.teacher_forcing:
            x, prev_y, labels = data
        else:
            x, labels = data

        n_s_feature_dim = self.config['enc_config']['n_s']
        n_h_feature_dim = self.config['enc_config']['n_h']

        if kwargs.get('use_datetime_index', False):  # during deindexification, first feature will be removed.
            n_s_feature_dim += 1
            n_h_feature_dim += 1
            idx = np.expand_dims(x[:, 1:, 0], axis=-1)   # extract the index from x
            if self.teacher_forcing:
                prev_y = np.concatenate([prev_y, idx], axis=2)  # insert index in prev_y

        if not self.config['drop_remainder']:
            s0 = np.zeros((x.shape[0], n_s_feature_dim))
            h0 = np.zeros((x.shape[0], n_h_feature_dim))
            x = [x, s0, h0]

        if self.verbosity > 0:
            print_something(x, "input_x")
            print_something(labels, "target")

        if self.teacher_forcing:
            return [x, prev_y], labels
        else:
            return x, labels

    def _transform_x(self, x, name):
        """transforms x and puts the transformer in config witht he key name
        for conformity we need to add feature names of initial states and their transformations
        will always be None."""
        feature_names = [
            self.input_features,
            [f"{i}" for i in range(self.enc_config['n_s'])],
            [f"{i}" for i in range(self.enc_config['n_h'])]
        ]
        transformation = [self.config['x_transformation'], None, None]
        return self._transform(x, name, transformation, feature_names)




# https://www.tensorflow.org/guide/keras/customizing_what_happens_in_fit

try:
    import tensorflow as tf
except ModuleNotFoundError:
    tf = None

#@tf.function
def train_step(keras_model, data):
    if int(''.join(tf.__version__.split('.')[0:2])) < 23:
        raise NotImplementedError(f"ignoring nan in labels can not be done in tf version {tf.__version__}")
    # Unpack the data. Its structure depends on your model and
    # on what you pass to `fit()`.
    x, y = data
    # TODO, in graph mode, masking of nans does not work

    with tf.GradientTape() as tape:
        y_pred = keras_model(x, training=True)  # Forward pass

        mask = tf.greater(y, 0.0)
        true_y = tf.boolean_mask(y, mask)
        pred_y = tf.boolean_mask(y_pred, mask)

        # Compute the loss value
        # (the loss function is configured in `compile()`)
        loss = keras_model.compiled_loss(true_y, pred_y, regularization_losses=keras_model.losses)

    # Compute gradients
    trainable_vars = keras_model.trainable_variables
    gradients = tape.gradient(loss, trainable_vars)
    # Update weights
    keras_model.optimizer.apply_gradients(zip(gradients, trainable_vars))
    # Update metrics (includes the metric that tracks the loss)
    keras_model.compiled_metrics.update_state(true_y, pred_y)
    # Return a dict mapping metric names to current value
    return {m.name: m.result() for m in keras_model.metrics}


#@tf.function
def test_step(keras_model, data):
    if int(''.join(tf.__version__.split('.')[0:2])) < 23:
        raise NotImplementedError(f"ignoring nan in labels can not be done in tf version {tf.__version__}")
    print('custom test_step')
    x, y = data

    y_pred = keras_model(x, training=False)  # compute predictions

    mask = tf.greater(y, 0.0)
    true_y = tf.boolean_mask(y, mask)
    pred_y = tf.boolean_mask(y_pred, mask)

    keras_model.compiled_loss(true_y, pred_y, regularization_losses=keras_model.losses)

    keras_model.compiled_metrics.update_state(true_y, pred_y)

    return {m.name: m.result() for m in keras_model.metrics}


__all__ = ["NN", "check_act_fn"]

from weakref import WeakKeyDictionary

try:
    from ai4water.tf_attributes import ACTIVATION_LAYERS, ACTIVATION_FNS, LAYERS, tf
except ModuleNotFoundError:
    tf = None

from . import backend as K

if K.BACKEND == 'tensorflow':
    from ai4water.tf_attributes import LAYERS, tf
else:
    try:  # maybe torch is also not available.
        from .models.torch import LAYERS
    except (ModuleNotFoundError, ImportError):
        LAYERS = {}


class AttributeNotSetYet:
    def __init__(self, func_name):
        self.data = WeakKeyDictionary()
        self.func_name = func_name

    def __get__(self, instance, owner):
        raise AttributeError("run the function {} first to get {}".format(self.func_name, self.name))

    def __set_name__(self, owner, name):
        self.name = name


class AttributeStore(object):

    """A class which will just make sure that attributes are set at its childs
    class level and not here. It's purpose is just to avoid cluttering of __init__
    method of its child classes. """
    def __init__(self):
        self._model = AttributeNotSetYet("`build` to build neural network")
        self.method = None
        self.en_densor_We = None
        self.en_LSTM_cell = None
        self.auto_enc_composite = None
        self.de_LSTM_cell = None
        self.de_densor_We = None
        self.scalers = {}
        self.is_training = False


class NN(AttributeStore):

    def __init__(self,
                 config: dict
                 ):

        self.config = config

        self.lookback = self.config['ts_args']['lookback']
        AttributeStore.__init__(self)

    @property
    def lookback(self):
        return self._lookback

    @lookback.setter
    def lookback(self, x):
        self._lookback = x

    def update_cache(self, cache: dict, key, value):
        if key in cache:
            raise ValueError("Duplicate input/output name found. The name '{}'"
                             " already exists as input/output for another layer"
                             .format(key))
        cache[key] = value
        return

    def deconstruct_lyr_args(self, lyr_name,  lyr_args) -> tuple:

        if not isinstance(lyr_args, dict):
            return lyr_args, None, None, None

        if callable(lyr_name):
            if hasattr(lyr_name, '__call__'):
                raise ValueError
            else:
                config = tf.keras.layers.Lambda(lambda x: lyr_name(x))
                inputs = lyr_args['inputs'] if 'inputs' in lyr_args else None
                outputs = lyr_args['outputs'] if 'outputs' in lyr_args else None
                call_args = lyr_args['call_args'] if 'call_args' in lyr_args else None

        elif 'config' not in lyr_args:
            if all([arg not in lyr_args for arg in ['inputs', 'outputs', 'call_args']]):
                config = lyr_args
                inputs = None
                outputs = None
                call_args = None
            else:  # todo, why we can't have inputs/outputs/call_args without config?
                raise ValueError(f"No config found for layer '{lyr_name}'")
        else:

            config = lyr_args['config']
            inputs = lyr_args['inputs'] if 'inputs' in lyr_args else None
            outputs = lyr_args['outputs'] if 'outputs' in lyr_args else None
            call_args = lyr_args['call_args'] if 'call_args' in lyr_args else None

        if tf is not None:
            if isinstance(config, tf.keras.layers.Lambda):
                config = tf.keras.layers.serialize(config)

        return config, inputs, outputs, call_args

    def check_lyr_config(self, lyr_name: str, config: dict):

        if callable(lyr_name):  # lyr_name is class
            if not isinstance(config, dict):
                config = {}
                args = [config]
            else:
                args = {}
            if 'name' not in config and K.BACKEND != 'pytorch':
                config['name'] = lyr_name.__name__
            config, activation = check_act_fn(config)

            return lyr_name.__name__, args, config, activation

        if callable(config):
            return lyr_name, [], config, None

        elif not isinstance(config, dict):
            args = [config]
            config = {}
        else:
            args = []

        if 'name' not in config and K.BACKEND != 'pytorch':
            config['name'] = lyr_name

        activation = None
        if "LAMBDA" not in lyr_name.upper():
            # for lambda layers, we don't need to check activation functions and layer names.
            config, activation = check_act_fn(config)

            # get keras/tensorflow layer compatible layer name
            lyr_name = self.get_layer_name(lyr_name)

        return lyr_name, args, config, activation

    def get_layer_name(self, lyr: str) -> str:

        layer_name = lyr.split('_')[0]
        if layer_name not in list(LAYERS.keys()) + list(ACTIVATION_LAYERS.keys()):
            raise ValueError(f"""
            The layer name '{lyr}' you specified, does not exist.
            Is this a user defined layer? If so, make sure your
            layer is being considered by AI4Water
                            """)

        return layer_name

    def get_and_set_attrs(self, layer):
        if layer.__class__.__name__ == "TemporalFusionTransformer":  # check the type without importing the layer
            # use layer name as there can be more than one layers from same class.
            setattr(self, f'{layer.name}_attentions', layer.attention_components)
        return


def check_act_fn(config: dict):
    """ it is possible that the config file does not have activation argument or activation is None"""
    activation = None
    if 'activation' in config:
        activation = config['activation']
    if activation is not None:
        assert isinstance(activation, str), f"unknown activation function {activation}"
        config['activation'] = ACTIVATION_FNS[activation]

    return config, activation


def get_call_args(lyr_inputs, lyr_cache, add_args, lyr_name):
    """ gets the additional call arguments for a layer. It is supposed that the
    call arguments are actually tensors/layers
    that have been created so far in the model including input layer. The call_args can be a list of inputs as well."""
    if isinstance(lyr_inputs, list):
        call_args = []
        for lyr_ins in lyr_inputs:
            if lyr_ins not in lyr_cache:
                raise ValueError("""No layer named '{}' currently exists in the model which can be fed
                                    as input to '{}' layer.""".format(lyr_ins, lyr_name))
            call_args.append(lyr_cache[lyr_ins])
    else:
        if lyr_inputs not in lyr_cache:
            raise ValueError(f"""
                                No layer named '{lyr_inputs}' currently exists in the model which
                                can be fed as input to '{lyr_name}' layer. Available layers are
                                {list(lyr_cache.keys())}
""")
        call_args = lyr_cache[lyr_inputs]

    return call_args, get_add_call_args(add_args, lyr_cache, lyr_name)


def get_add_call_args(add_args, lyr_cache, lyr_name):
    additional_args = {}
    if add_args is not None:
        assert isinstance(add_args, dict), "call_args to layer '{}' must be provided as dictionary".format(lyr_name)
        for arg_name, arg_val in add_args.items():
            if isinstance(arg_val, str):
                if arg_val not in lyr_cache:
                    raise NotImplementedError("The value {} for additional call argument {} to '{}'"
                                              " layer not understood".format(arg_val, arg_name, lyr_name))
                additional_args[arg_name] = lyr_cache[arg_val]

            elif isinstance(arg_val, list):
                # the additional argument is a list of tensors, get all of them from lyr_cache
                add_arg_val_list = []
                for arg in arg_val:
                    assert isinstance(arg, str)
                    add_arg_val_list.append(lyr_cache[arg])

                additional_args[arg_name] = add_arg_val_list

            elif isinstance(arg_val, bool) or arg_val is None:
                additional_args[arg_name] = arg_val

            else:
                raise NotImplementedError("The value `{}` for additional call argument {} to '{}'"
                                          " layer not understood".format(arg_val, arg_name, lyr_name))

    return additional_args



import os
from typing import Union, Callable

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

try:
    import plotly
except ModuleNotFoundError:
    plotly = None

from .plotting_tools import save_or_show, to_1d_array


class Plot(object):

    def __init__(self, path=None, backend='plotly', save=True, dpi=300):
        self.path = path
        self.backend = backend
        self.save = save
        self.dpi = dpi

    @property
    def backend(self):
        return self._backend

    @backend.setter
    def backend(self, x):

        _backend = x
        assert x in ['plotly', 'matplotlib'], f"unknown backend {x}. Allowed values are `plotly` and `matplotlib`"

        if x == 'plotly':
            if plotly is None:
                _backend = 'matplotlib'

        self._backend = _backend

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, x):
        if x is None:
            x = os.getcwd()
        self._path = x

    def save_or_show(self, save: bool = None, fname=None, where='', dpi=None, bbox_inches='tight',
                     close=True, show=False):

        if save is None:
            save = self.save

        if dpi is None:
            dpi = self.dpi

        return save_or_show(self.path, save, fname, where, dpi, bbox_inches, close, show=show)


def linear_model(
        model_name: str,
        inputs,
        target
):
    import sklearn
    from ai4water.backend import get_attributes

    models = get_attributes(sklearn, "linear_model", case_sensitive=True)
    if model_name not in models:
        raise ValueError(f"Can not find {model_name} in sklearn.linear_model")
    model = models[model_name]
    reg = model().fit(inputs, target)

    return reg.predict(inputs)


def murphy_diagram(
        observed: Union[list, np.ndarray, pd.Series, pd.DataFrame],
        predicted: Union[list, np.ndarray, pd.Series, pd.DataFrame],
        reference: Union[list, np.ndarray, pd.Series, pd.DataFrame] = None,
        reference_model: Union[str, Callable] = None,
        inputs=None,
        plot_type: str = "scores",
        xaxis: str = "theta",
        ax: plt.Axes = None,
        line_colors: tuple = None,
        fill_color: str = "lightgray",
        show: bool = True
) -> plt.Axes:
    """Murphy diagram as introducted by Ehm_ et al., 2015
     and illustrated by Rob Hyndman_

    Arguments:
        observed:
            observed or true values
        predicted:
             model's prediction
        reference:
             reference prediction
        reference_model:
             The model for reference prediction. Only relevent if `reference` is
             None and `plot_type` is `diff`. It can be callable or a string. If it is a
             string, then it can be any model name from sklearn.linear_model_
        inputs:
             inputs for reference model. Only relevent if `reference_model` is not
             None and `plot_type` is `diff`
        plot_type:
             either of `scores` or `diff`
        xaxis:
             either of `theta` or `time`
        ax:
             the axis to use for plotting
        line_colors:
             colors of line
        fill_color:
             color to fill confidence interval
        show:
             whether to show the plot or not

    Returns:
         matplotlib axes

    Example:
        >>> import numpy as np
        >>> from ai4water.utils.visualizations import murphy_diagram
        >>> yy = np.random.randint(1, 1000, 100)
        >>> ff1 = np.random.randint(1, 1000, 100)
        >>> ff2 = np.random.randint(1, 1000, 100)
        >>> murphy_diagram(yy, ff1, ff2)
        ...
        >>> murphy_diagram(yy, ff1, ff2, plot_type="diff")

    .. _Ehm:
        https://arxiv.org/pdf/1503.08195.pdf

    .. _Hyndman:
        https://robjhyndman.com/hyndsight/murphy-diagrams/

    .. _sklearn.linear_model:
        https://scikit-learn.org/stable/modules/classes.html#module-sklearn.linear_model
    """
    assert plot_type in ("scores", "diff")
    assert xaxis in ("theta", "time")

    y = to_1d_array(observed)
    f1 = to_1d_array(predicted)

    if reference is None:
        if plot_type == "diff":
            assert reference_model is not None
            if callable(reference_model):
                reference = reference_model(inputs)
            else:
                assert inputs is not None, f"You must specify the inputs for {reference_model}"
                reference = linear_model(reference_model, inputs, predicted)
            f2 = to_1d_array(reference)
        else:
            f2 = None
    else:
        f2 = to_1d_array(reference)

    line_colors = line_colors or ["dimgrey", "tab:orange"]

    n = len(y)
    _min, _max = np.nanmin(np.hstack([y, f1, f2])), np.nanmax(np.hstack([y, f1, f2]))
    tmp = _min-0.2*(_max-_min), _max+0.2*(_max-_min)

    theta = np.linspace(tmp[0], tmp[1], 501)

    s1 = np.full((501, n), np.nan)
    s2 = np.full((501, n), np.nan)

    max1 = np.maximum(f1, y)
    max2 = np.maximum(f2, y)

    min1 = np.minimum(f1, y)
    min2 = np.minimum(f2, y)

    for j in range(n):

        s1[:, j] = abs(y[j]-theta) * (max1[j] > theta) * (min1[j] <= theta)
        s2[:, j] = abs(y[j] - theta) * (max2[j] > theta) * (min2[j] <= theta)

    # grab the axes
    if ax is None:
        ax = plt.gca()

    if xaxis == "theta":
        s1ave, s2ave = _data_for_theta(s1, s2)
    else:
        raise NotImplementedError

    if plot_type == "scores":
        _plot_scores(theta, s1ave, s2ave, ax, line_colors)
        ax.set_ylabel("Empirical Scores", fontsize=16)
    else:
        _plot_diff(theta, s1, s2, n, ax, line_colors[0], fill_color)
        ax.set_ylabel("Difference in scores", fontsize=16)

    ax.set_xlabel(xaxis, fontsize=16)

    if show:
        plt.show()

    return ax


def last_nonzero(arr, axis, invalid_val=-1):
    mask = arr != 0
    val = arr.shape[axis] - np.flip(mask, axis=axis).argmax(axis=axis) - 1
    return np.where(mask.any(axis=axis), val, invalid_val)


def _plot_diff(theta, s1, s2, n, ax, line_color="black", fill_color="lightgray"):

    se = np.std(s1-s2)/np.sqrt(n)

    diff = np.mean(s1-s2, axis=1)

    upper = diff + 1.96 * se
    lower = diff - 1.96 * se

    ax.plot(theta, diff, color=line_color)

    # first_nonzero occurence
    st = (diff != 0).argmax(axis=0)
    en = last_nonzero(diff, axis=0).item()

    ax.fill_between(theta[st:en], upper[st:en], lower[st:en],  # alpha=0.2,
                    color=fill_color)

    return ax


def fdc_plot(
        sim: Union[list, np.ndarray, pd.Series, pd.DataFrame],
        obs: Union[list, np.ndarray, pd.Series, pd.DataFrame],
        ax: plt.Axes = None,
        legend: bool = True,
        xlabel: str = "Exceedence [%]",
        ylabel: str = "Flow",
        show: bool = True
) -> plt.Axes:
    """Plots flow duration curve

    Arguments:
        sim:
            simulated flow
        obs:
            observed flow
        ax:
            axis on which to plot
        legend:
            whether to apply legend or not
        xlabel:
            label to set on x-axis. set to None for no x-label
        ylabel:
            label to set on y-axis
        show:
            whether to show the plot or not

    Returns:
        matplotlib axes

    Example:
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from ai4water.utils.visualizations import fdc_plot
        >>> simulated = np.random.random(100)
        >>> observed = np.random.random(100)
        >>> fdc_plot(simulated, observed)
        >>> plt.show()
    """

    sim = to_1d_array(sim)
    obs = to_1d_array(obs)

    sort_obs = np.sort(sim)[::-1]
    exceedence_obs = np.arange(1., len(sort_obs)+1) / len(sort_obs)
    sort_sim = np.sort(obs)[::-1]
    exceedence_sim = np.arange(1., len(sort_sim)+1) / len(sort_sim)

    if ax is None:
        ax = plt.gca()

    ax.plot(exceedence_obs * 100, sort_obs, color='b', label="Observed")
    ax.plot(exceedence_sim * 100, sort_sim, color='r', label="Simulated")

    if legend:
        ax.legend()

    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if show:
        plt.show()

    return ax


def _plot_scores(theta, s1ave, s2ave, ax, line_colors):

    ax.plot(theta, s1ave, color=line_colors[0])
    ax.plot(theta, s2ave, color=line_colors[1])

    return ax


def _data_for_time(s1, s2):
    s1ave, s2ave = np.mean(s1, axis=0), np.mean(s2, axis=0)

    return s1ave, s2ave


def _data_for_theta(s1, s2):
    return np.mean(s1, axis=1), np.mean(s2, axis=1)

def init_subplots(width=None, height=None, nrows=1, ncols=1, **kwargs):
    """Initializes the fig for subplots"""
    plt.close('all')
    fig, ax = plt.subplots(nrows=nrows, ncols=ncols, **kwargs)
    if width is not None:
        fig.set_figwidth(width)
    if height is not None:
        fig.set_figheight(height)
    return fig, ax


import copy
import os
import json
import pprint
import datetime
import warnings
from typing import Union
from shutil import rmtree
from collections import OrderedDict
from typing import Tuple, List
import collections.abc as collections_abc

import scipy
import numpy as np
import pandas as pd
from easy_mpl import imshow
import matplotlib.pyplot as plt
from scipy.stats import skew, kurtosis, variation, gmean, hmean

try:
    import wrapt
except ModuleNotFoundError:
    wrapt = None


MATRIC_TYPES = {
    "r2": "max",
    "nse": "max",
    "r2_score": "max",
    "kge": "max",
    "corr_coeff": "max",
    'accuracy': "max",
    'f1_score': 'max',
    "mse": "min",
    "rmse": "min",
    "mape": "min",
    "nrmse": "min",
}

ERROR_LABELS = {
    'r2': "$R^{2}$",
    'nse': 'NSE',
    'rmse': 'RMSE',
    'mse': 'MSE',
    'msle': 'MSLE',
    'nrmse': 'Normalized RMSE',
    'mape': 'MAPE',
    'r2_score': "$R^{2}$ Score"
}

def reset_seed(seed: Union[int, None], os=None, random=None, np=None, tf=None, torch=None):
    """
    Sets the random seed for a given module if the module is not None
    Arguments:
        seed : Value of seed to set. If None, then it means we don't wan't to set
        the seed.
        os : alias for `os` module of python
        random : alias for `random` module of python
        np : alias for `numpy` module
        tf : alias for `tensorflow` module.
        torch : alias for `pytorch` module.
        """
    if seed:
        if np:
            np.random.seed(seed)

        if random:
            random.seed(seed)

        if os:
            os.environ['PYTHONHASHSEED'] = str(seed)

        if tf:
            if int(tf.__version__.split('.')[0]) == 1:
                tf.compat.v1.random.set_random_seed(seed)
            elif int(tf.__version__.split('.')[0]) > 1:
                tf.random.set_seed(seed)

        if torch:
            torch.manual_seed(seed)
            torch.cuda.manual_seed(seed)
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = True
    return


def maybe_create_path(prefix=None, path=None):
    if path is None:
        save_dir = dateandtime_now()
        model_dir = os.path.join(os.getcwd(), "results")

        if prefix:
            model_dir = os.path.join(model_dir, prefix)

        save_dir = os.path.join(model_dir, save_dir)
    else:
        save_dir = path

    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    for _dir in ['weights']:
        if not os.path.exists(os.path.join(save_dir, _dir)):
            os.makedirs(os.path.join(save_dir, _dir))

    return save_dir


def dateandtime_now() -> str:
    """
    Returns the datetime in following format as string
    YYYYMMDD_HHMMSS
    """
    jetzt = datetime.datetime.now()
    dt = ''
    for time in ['year', 'month', 'day', 'hour', 'minute', 'second']:
        _time = str(getattr(jetzt, time))
        if len(_time) < 2:
            _time = '0' + _time
        if time == 'hour':
            _time = '_' + _time
        dt += _time
    return dt


def dict_to_file(path, config=None, errors=None, indices=None, others=None, name=''):

    sort_keys = True
    if errors is not None:
        suffix = dateandtime_now()
        fpath = path + "/errors_" + name + suffix + ".json"
        # maybe some errors are not json serializable.
        for er_name, er_val in errors.items():
            if "int" in er_val.__class__.__name__:
                errors[er_name] = int(er_val)
            elif "float" in er_val.__class__.__name__:
                errors[er_name] = float(er_val)

        data = errors
    elif config is not None:
        fpath = path + "/config.json"
        data = config
        sort_keys = False
    elif indices is not None:
        fpath = path + "/indices.json"
        data = indices
    else:
        assert others is not None
        data = others
        fpath = path

    if 'config' in data:
        if data['config'].get('model', None) is not None:
            model = data['config']['model']
            if 'layers' not in model:  # because ML args which come algorithms may not be of json serializable.
                model = Jsonize(model)()
                data['config']['model'] = model

    with open(fpath, 'w') as fp:
        json.dump(data, fp, sort_keys=sort_keys, indent=4, cls=JsonEncoder)

    return


def check_min_loss(epoch_losses, epoch, msg: str, save_fg: bool, to_save=None):
    epoch_loss_array = epoch_losses[:-1]

    current_epoch_loss = epoch_losses[-1]

    if len(epoch_loss_array) > 0:
        min_loss = np.min(epoch_loss_array)
    else:
        min_loss = current_epoch_loss

    if np.less(current_epoch_loss, min_loss):
        msg = msg + "    {:10.5f} ".format(current_epoch_loss)

        if to_save is not None:
            save_fg = True
    else:
        msg = msg + "              "

    return msg, save_fg


def check_kwargs(**kwargs):

    mode = "ML"
    if kwargs.get('model', None) is not None:
        model = kwargs['model']
        if isinstance(model, dict):
            if 'layers' in model:
                is_custom_model=False
                model_name = None
                mode="DL"
            else:
                assert len(model)==1
                _model = list(model.keys())[0]
                if isinstance(_model, str):
                    model_name = _model
                    is_custom_model = False
                elif hasattr(_model, '__call__'):   # uninitiated class
                    check_attributes(_model, ['fit', 'predict', '__init__'])
                    model_name = _model.__name__
                    is_custom_model = True
                else:  # custom class is already initiated
                    check_attributes(_model, ['fit', 'predict'])
                    is_custom_model = True
                    model_name = _model.__class__.__name__

        # for case when model='randomforestregressor'
        elif isinstance(model, str):
            kwargs['model'] = {model: {}}
            is_custom_model = False
            model_name = model

        elif hasattr(model, '__call__'):  # uninitiated class
            check_attributes(model, ['fit', 'predict', '__init__'])
            model_name = model.__name__
            is_custom_model = True
            kwargs['model'] = {model: {}}

        else:
            check_attributes(model, ['fit', 'predict'])
            is_custom_model = True
            model_name = model.__class__.__name__
            kwargs['model'] = {model: {}}

        if mode=="ML":
            if "batches" not in kwargs:  # for ML, default batches will be 2d unless the user specifies otherwise.
                kwargs["batches"] = "2d"

            if "ts_args" not in kwargs:
                kwargs["ts_args"] = {'lookback': 1,
                                     'forecast_len': 1,
                                     'forecast_step': 0,
                                     'known_future_inputs': False,
                                     'input_steps': 1,
                                     'output_steps': 1}
    else:
        is_custom_model = False
        model_name = None

    if is_custom_model:
        if 'mode' not in kwargs:
            raise ValueError("""your must provide 'mode' keyword either as 
            mode='regression' or mode='classification' for custom models""")

    return kwargs, model_name, is_custom_model


class make_model(object):

    def __init__(self, **kwargs):

        self.config, self.data_config, self.opt_paras, self.orig_model = _make_model(**kwargs)


def process_io(**kwargs):

    input_features = kwargs.get('input_features', None)
    output_features = kwargs.get('output_features', None)

    if isinstance(input_features, str):
        input_features = [input_features]
    if isinstance(output_features, str):
        output_features = [output_features]

    kwargs['input_features'] = input_features
    kwargs['output_features'] = output_features
    return kwargs


def _make_model(**kwargs):
    """
    This functions fills the default arguments needed to run all the models.
    All the input arguments can be overwritten
    by providing their name.
    :return
      nn_config: `dict`, contais parameters to build and train the neural network
        such as `layers`
      data_config: `dict`, contains parameters for data preparation/pre-processing/post-processing etc.
    """

    kwargs = process_io(**kwargs)

    kwargs, model_name, is_custom_model = check_kwargs(**kwargs)

    model = kwargs.get('model', None)
    def_cat = None

    if model is not None:
        if 'layers' in model:
            def_cat = "DL"
            # for DL, the default mode case will be regression
        else:
            def_cat = "ML"

    accept_additional_args = False
    if 'accept_additional_args' in kwargs:
        accept_additional_args = kwargs.pop('accept_additional_args')

    model_args = {

        'model': {'type': dict, 'default': None, 'lower': None, 'upper': None, 'between': None},
        # can be None or any of the method defined in ai4water.utils.transformatinos.py
        'x_transformation': {"type": [str, type(None), dict, list], "default": None, 'lower': None,
                             'upper': None, 'between': None},
        'y_transformation': {"type": [str, type(None), dict, list], "default": None, 'lower': None,
                             'upper': None, 'between': None},
        # for auto-encoders
        'composite':    {'type': bool, 'default': False, 'lower': None, 'upper': None, 'between': None},
        'lr':           {'type': float, 'default': 0.001, 'lower': None, 'upper': None, 'between': None},
        # can be any of valid keras optimizers https://www.tensorflow.org/api_docs/python/tf/keras/optimizers
        'optimizer':    {'type': str, 'default': 'adam', 'lower': None, 'upper': None, 'between': None},
        'loss':         {'type': [str, 'callable'], 'default': 'mse', 'lower': None, 'upper': None, 'between': None},
        'quantiles':    {'type': list, 'default': None, 'lower': None, 'upper': None, 'between': None},
        'epochs':       {'type': int, 'default': 14, 'lower': None, 'upper': None, 'between': None},
        'min_val_loss': {'type': float, 'default': 0.0001, 'lower': None, 'upper': None, 'between': None},
        'patience':     {'type': int, 'default': 100, 'lower': None, 'upper': None, 'between': None},
        'shuffle':      {'type': bool, 'default': True, 'lower': None, 'upper': None, 'between': None},
        # to save the best models using checkpoints
        'save_model':   {'type': bool, 'default': True, 'lower': None, 'upper': None, 'between': None},
        # used for cnn_lst structure
        'subsequences': {'type': int, 'default': 3, 'lower': 2, "upper": None, "between": None},

        'backend':       {'type': None, 'default': 'tensorflow', 'lower': None, 'upper': None,
                          'between': ['tensorflow', 'pytorch']},
        # buffer_size is only relevant if 'val_data' is same and shuffle is true.
        # https://www.tensorflow.org/api_docs/python/tf/data/Dataset#shuffle
        # It is used to shuffle tf.Dataset of training data.
        'buffer_size': {'type': int, 'default': 100, 'lower': None, 'upper': None, 'between': None},
        # comes handy if we want to skip certain batches from last
        'batches_per_epoch': {"type": int, "default": None, 'lower': None, 'upper': None, 'between': None},
        # https://www.tensorflow.org/api_docs/python/tf/keras/Model#fit
        'steps_per_epoch': {"type": int, "default": None, 'lower': None, 'upper': None, 'between': None},
        # can be string or list of strings such as 'mse', 'kge', 'nse', 'pbias'
        'metrics': {"type": list, "default": ['nse'], 'lower': None, 'upper': None, 'between': None},
        # todo, is it  redundant?
        # If the model takes one kind of input_features that is it consists of
        # only 1 Input layer, then the shape of the batches
        # will be inferred from the Input layer but for cases, the model takes more than 1 Input, then there can be two
        # cases, either all the input_features are of same shape or they
        # are not. In second case, we should overwrite `train_paras`
        # method. In former case, define whether the batches are 2d or 3d. 3d means it is for an LSTM and 2d means it is
        # for Dense layer.
        'batches': {"type": str, "default": '3d', 'lower': None, 'upper': None, 'between': ["2d", "3d"]},
        'prefix': {"type": str, "default": None, 'lower': None, 'upper': None, 'between': None},
        'path': {"type": str, "default": None, 'lower': None, 'upper': None, 'between': None},
        'kmodel': {'type': None, "default": None, 'lower': None, 'upper': None, 'between': None},
        'cross_validator': {'default': None, 'between': ['LeaveOneOut', 'kfold']},
        'wandb_config': {'type': dict, 'default': None, 'between': None},
        'val_metric': {'type': str, 'default': None},
        'model_name_': {'default': None},
        'is_custom_model_': {"default": None},
    }

    data_args = {
        # if the shape of last batch is smaller than batch size and if we
        # want to skip this last batch, set following to True.
        # Useful if we have fixed batch size in our model but the number of samples is not fully divisble by batch size
        'drop_remainder': {"type": bool, "default": False, 'lower': None, 'upper': None, 'between': [True, False]},
        'category': {'type': str, 'default': def_cat, 'lower': None, 'upper': None, 'between': ["ML", "DL"]},
        'mode': {'type': str, 'default': None, 'lower': None, 'upper': None,
                 'between': ["regression", "classification"]},
        'batch_size':        {"type": int,   "default": 32, 'lower': None, 'upper': None, 'between': None},
        'split_random': {'type': bool, 'default': False, 'between': [True, False]},
        # fraction of data to be used for validation
        'val_fraction':      {"type": float, "default": 0.2, 'lower': None, 'upper': None, 'between': None},
        # the following argument can be set to 'same' for cases if you want to use same data as validation as well as
        # test data. If it is 'same', then same fraction/amount of data will be used for validation and test.
        # If this is not string and not None, this will overwite `val_fraction`
        'indices':          {"type": dict,  "default": None, 'lower': None, 'upper': None, 'between': ["same", None]},
        # fraction of data to be used for test
        'train_fraction':     {"type": float, "default": 0.7, 'lower': None, 'upper': None, 'between': None},
        # write the data/batches as hdf5 file
        'save':        {"type": bool,  "default": False, 'lower': None, 'upper': None, 'between': None},
        'allow_nan_labels':       {"type": int,  "default": 0, 'lower': 0, 'upper': 2, 'between': None},

        'nan_filler':        {"type": None, "default": None, "lower": None, "upper": None, "between": None},

        # for reproducability
        'seed':              {"type": None, "default": 313, 'lower': None, 'upper': None, 'between': None},
        # input features in data_frame
        'input_features':            {"type": None, "default": None, 'lower': None, 'upper': None, 'between': None},
        # column in dataframe to bse used as output/target
        'output_features':           {"type": None, "default": None, 'lower': None, 'upper': None, 'between': None},
        # tuple of tuples where each tuple consits of two integers, marking the start and end
        # of interval. An interval here
        # means chunk/rows from the input file/dataframe to be skipped when when preparing
        # data/batches for NN. This happens
        # when we have for example some missing values at some time in our data.
        # For further usage see `examples/using_intervals`
        "intervals":         {"type": None, "default": None, 'lower': None, 'upper': None, 'between': None},
        'verbosity':         {"type": int, "default": 1, 'lower': None, 'upper': None, 'between': None},
        'teacher_forcing': {'type': bool, 'default': False, 'lower': None, 'upper': None, 'between': [True, False]},
        'dataset_args': {'type': dict, 'default': {}},
        'ts_args': {"type": dict, "default": {'lookback': 1,
                                              'forecast_len': 1,
                                              'forecast_step': 0,
                                              'known_future_inputs': False,
                                              'input_steps': 1,
                                              'output_steps': 1}}
    }

    model_config = {key: val['default'] for key, val in model_args.items()}

    config = {key: val['default'] for key, val in data_args.items()}

    opt_paras = {}
    # because there are two kinds of hpos which can be optimized
    # some can be in model config and others are in main config
    original_other_conf = {}
    original_mod_conf = {}

    for key, val in kwargs.items():
        arg_name = key.lower()  # todo, why this?

        if val.__class__.__name__ in ['Integer', "Real", "Categorical"]:
            opt_paras[key] = val
            val2 = val
            val = jsonize(val.rvs(1)[0])

            val2.name = key
            original_other_conf[key] = val2

        if key == 'model':
            val, _opt_paras, original_mod_conf = find_opt_paras_from_model_config(val)
            opt_paras.update(_opt_paras)

        if key == 'ts_args':
            val, _opt_paras = find_opt_paras_from_ts_args(val)
            opt_paras.update(_opt_paras)

        if arg_name in model_config:
            update_dict(arg_name, val, model_args, model_config)

        elif arg_name in config:
            update_dict(arg_name, val, data_args, config)

        elif arg_name in ['x_transformer_', 'y_transformer_', 'val_x_transformer_', 'val_y_transformer_']:
            pass

        # config may contain additional user defined args which will not be checked
        elif not accept_additional_args:
            raise ValueError(f"Unknown keyworkd argument '{key}' provided")
        else:
            config[key] = val

    if config['allow_nan_labels'] > 0:
        assert 'layers' in model_config['model'], f"""
The model appears to be deep learning based because 
the argument `model` does not have layers. But you are
allowing nan labels in the targets.
However, `allow_nan_labels` should be > 0 only for deep learning models
"""

    config.update(model_config)

    if isinstance(config['input_features'], dict):
        for data in [config['input_features'], config['output_features']]:
            for k, v in data.items():
                assert isinstance(v, list), f"{k} is of type {v.__class__.__name__} but it must of of type list"

    _data_config = {}
    for key, val in config.items():
        if key in data_args:
            _data_config[key] = val

    config['model_name_'] = model_name
    config['is_custom_model_'] = is_custom_model

    return config, _data_config, opt_paras, {'model': original_mod_conf, 'other': original_other_conf}


def update_dict(key, val, dict_to_lookup, dict_to_update):
    """Updates the dictionary with key, val if the val is of type dtype."""
    dtype = dict_to_lookup[key].get('type', None)
    low = dict_to_lookup[key].get('lower', None)
    up = dict_to_lookup[key].get('upper', None)
    between = dict_to_lookup[key].get('between', None)

    if dtype is not None:
        if isinstance(dtype, list):
            val_type = type(val)
            if 'callable' in dtype:
                if callable(val):
                    pass

            elif val_type not in dtype:
                raise TypeError("{} must be any of the type {} but it is of type {}"
                                .format(key, dtype, val.__class__.__name__))
        elif not isinstance(val, dtype):
            # the default value may be None which will be different than dtype
            if val != dict_to_lookup[key]['default']:
                raise TypeError(f"{key} must be of type {dtype} but it is of type {val.__class__.__name__}")

    if isinstance(val, int) or isinstance(val, float):
        if low is not None:
            if val < low:
                raise ValueError(f"The value '{val}' for '{key}' must be greater than '{low}'")
        if up is not None:
            if val > up:
                raise ValueError(f"The value '{val} for '{key} must be less than '{up}'")

    if isinstance(val, str):
        if between is not None:
            if val not in between:
                raise ValueError(f"Unknown value '{val}' for '{key}'. It must be one of '{between}'")

    dict_to_update[key] = val
    return


def deepcopy_dict_without_clone(d: dict) -> dict:
    """makes deepcopy of a dictionary without cloning it"""
    assert isinstance(d, dict)

    new_d = {}
    for k, v in d.items():
        if isinstance(v, dict):
            new_d[k] = deepcopy_dict_without_clone(v)
        elif hasattr(v, '__len__'):
            new_d[k] = v[:]
        else:
            new_d[k] = copy.copy(v)
    return new_d


def find_opt_paras_from_ts_args(ts_args:dict)->tuple:
    opt_paras = {}
    new_ts_args = {'lookback': 15,
                   'forecast_len': 1,
                   'forecast_step': 0,
                   'known_future_inputs': False,
                   'input_steps': 1,
                   'output_steps': 1}
    new_ts_args.update(ts_args)

    for k,v in ts_args.items():

        if v.__class__.__name__ in ['Integer', 'Real', 'Categorical']:
            if v.name is None or v.name.startswith("integer_") or v.name.startswith("real_"):
                v.name = k
            opt_paras[k] = v
            v = v.rvs(1)[0]
            new_ts_args[k] = v

    return new_ts_args, opt_paras

def find_opt_paras_from_model_config(
        config: Union[dict, str, None]
) -> Tuple[Union[dict, None, str], dict, Union[dict, str, None]]:

    opt_paras = {}

    if config is None or isinstance(config, str):
        return config, opt_paras, config

    assert isinstance(config, dict) and len(config) == 1

    if 'layers' in config:
        original_model_config, _ = process_config_dict(deepcopy_dict_without_clone(config['layers']), False)

        # it is a nn based model
        new_lyrs_config, opt_paras = process_config_dict(config['layers'])
        new_model_config = {'layers': new_lyrs_config}

    else:
        # it is a classical ml model
        _ml_config = {}
        ml_config: dict = list(config.values())[0]
        model_name = list(config.keys())[0]

        original_model_config, _ = process_config_dict(copy.deepcopy(config[model_name]), False)

        for k, v in ml_config.items():

            if v.__class__.__name__ in ['Integer', 'Real', 'Categorical']:

                if v.name is None or v.name.startswith("integer_") or v.name.startswith("real_"):
                    v.name = k
                opt_paras[k] = v
                v = v.rvs(1)[0]

            _ml_config[k] = v
        val = _ml_config
        new_model_config = {model_name: val}

    return new_model_config, opt_paras, original_model_config


def process_config_dict(config_dict: dict, update_initial_guess=True):
    """From a dicitonary defining structure of neural networks, this function
    finds out which are hyperparameters from them"""

    assert isinstance(config_dict, dict)

    opt_paras = {}

    def pd(d):
        for k, v in d.items():
            if isinstance(v, dict) and len(v) > 0:
                pd(v)
            elif v.__class__.__name__ in ["Integer", "Real", "Categorical"]:
                if v.name is None or v.name.startswith("integer_") or v.name.startswith("real_"):
                    v.name = k

                if v.name in opt_paras:
                    raise ValueError("Hyperparameters with duplicate name found. A hyperparameter to be "
                                     f"optimized with name '{v.name}' already exists")
                opt_paras[v.name] = v
                if update_initial_guess:
                    x0 = jsonize(v.rvs(1)[0])  # get initial guess
                    d[k] = x0  # inplace change of dictionary
                else:  # we most probably have updated the name, so doing inplace change
                    d[k] = v
        return

    pd(config_dict)
    return config_dict, opt_paras


def update_model_config(config: dict, suggestions):
    """returns the updated config if config contains any parameter from suggestions."""
    cc = copy.deepcopy(config)

    def update(c):
        for k, v in c.items():
            if isinstance(v, dict):
                update(v)
            elif v.__class__.__name__ in ["Integer", "Real", "Categorical"]:
                c[k] = suggestions[v.name]
        return

    update(cc)

    return cc


def to_datetime_index(idx_array, fmt='%Y%m%d%H%M') -> pd.DatetimeIndex:
    """ converts a numpy 1d array into pandas DatetimeIndex type."""

    if not isinstance(idx_array, np.ndarray):
        raise TypeError

    idx = pd.to_datetime(idx_array.astype(str), format=fmt)
    idx.freq = pd.infer_freq(idx)
    return idx


class Jsonize(object):
    """Converts the objects to json compatible format i.e to native python types.
    If the object is sequence then each member of the sequence is checked and
    converted if needed. Same goes for nested sequences like lists of lists
    or list of dictionaries.

    Examples:
    ---------
    >>>import numpy as np
    >>>from ai4water.utils.utils import Jsonize
    >>>a = np.array([2.0])
    >>>b = Jsonize(a)(a)
    >>>type(b)  # int
    """
    # TODO, repeating code in __call__ and stage2
    # TODO, stage2 not considering tuple

    def __init__(self, obj):
        self.obj = obj

    def __call__(self):
        """Serializes one object"""
        if 'int' in self.obj.__class__.__name__:
            return int(self.obj)
        if 'float' in self.obj.__class__.__name__:
            return float(self.obj)

        if isinstance(self.obj, dict):
            return {k: self.stage2(v) for k, v in self.obj.items()}

        if hasattr(self.obj, '__len__') and not isinstance(self.obj, str):
            return [self.stage2(i) for i in self.obj]

        # if obj is a python 'type'
        if type(self.obj).__name__ == type.__name__:
            return self.obj.__name__

        if isinstance(self.obj, collections_abc.Mapping):
            return dict(self.obj)

        if self.obj is Ellipsis:
            return {'class_name': '__ellipsis__'}

        if wrapt and isinstance(self.obj, wrapt.ObjectProxy):
            return self.obj.__wrapped__

        return str(self.obj)

    def stage2(self, obj):
        """Serializes one object"""
        if any([isinstance(obj, _type) for _type in [bool, set, type(None)]]) or callable(obj):
            return obj

        if 'int' in obj.__class__.__name__:
            return int(obj)

        if 'float' in obj.__class__.__name__:
            return float(obj)

        # tensorflow tensor shape
        if obj.__class__.__name__ == 'TensorShape':
            return obj.as_list()

        if isinstance(obj, dict):  # iterate over obj until it is a dictionary
            return {k: self.stage2(v) for k, v in obj.items()}

        if hasattr(obj, '__len__') and not isinstance(obj, str):
            if len(obj) > 1:  # it is a list like with length greater than 1
                return [self.stage2(i) for i in obj]
            elif isinstance(obj, list) and len(obj) > 0:  # for cases like obj is [np.array([1.0])] -> [1.0]
                return [self.stage2(obj[0])]
            elif len(obj) == 1:  # for cases like obj is np.array([1.0])
                if isinstance(obj, list) or isinstance(obj, tuple):
                    return obj  # for cases like (1, ) or [1,]
                return self.stage2(obj[0])
            else:  # when object is []
                return obj

        # if obj is a python 'type'
        if type(obj).__name__ == type.__name__:
            return obj.__name__

        if obj is Ellipsis:
            return {'class_name': '__ellipsis__'}

        if wrapt and isinstance(obj, wrapt.ObjectProxy):
            return obj.__wrapped__

        # last solution, it must be of of string type
        return str(obj)


def jsonize(obj):
    """functional interface to `Jsonize` class"""
    return Jsonize(obj)()


def make_hpo_results(opt_dir, metric_name='val_loss') -> dict:
    """Looks in opt_dir and saves the min val_loss with the folder name"""
    results = {}
    for folder in os.listdir(opt_dir):
        fname = os.path.join(os.path.join(opt_dir, folder), 'losses.csv')

        if os.path.exists(fname):
            df = pd.read_csv(fname)

            if 'val_loss' in df:
                min_val_loss = round(float(np.nanmin(df[metric_name])), 6)
                results[min_val_loss] = {'folder': os.path.basename(folder)}
    return results


def find_best_weight(w_path: str, best: str = "min", ext: str = ".hdf5", epoch_identifier: int = None):
    """Given weights in w_path, find the best weight.
    if epoch_identifier is given, it will be given priority to find best_weights
    The file_names are supposed in following format FileName_Epoch_Error.ext

    Note: if we are monitoring more than two metrics whose desired behaviour
        is opposite to each other then this method does not work as desired. However
        this can be avoided by specifying `epoch_identifier`.
    """
    assert best in ['min', 'max']
    all_weights = os.listdir(w_path)

    if len(all_weights) == 1:
        return all_weights[0]

    losses = {}
    for w in all_weights:
        wname = w.split(ext)[0]
        try:
            val_loss = str(float(wname.split('_')[2]))  # converting to float so that trailing 0 is removed
        except ValueError as e:
            raise ValueError(f"while trying to find best weight in {w_path} with {best} and"
                             f" {ext} and {epoch_identifier}"
                             f" encountered following error \n{e}")
        losses[val_loss] = {'loss': wname.split('_')[2], 'epoch': wname.split('_')[1]}

    best_weight = None
    if epoch_identifier:
        for v in losses.values():
            if str(epoch_identifier) in v['epoch']:
                best_weight = f"weights_{v['epoch']}_{v['loss']}.hdf5"
                break

    else:
        loss_array = np.array([float(l) for l in losses.keys()])
        if len(loss_array) == 0:
            return None
        best_loss = getattr(np, best)(loss_array)
        best_weight = f"weights_{losses[str(best_loss)]['epoch']}_{losses[str(best_loss)]['loss']}.hdf5"

    return best_weight


def remove_all_but_best_weights(w_path, best: str = "min", ext: str = ".hdf5"):
    """removes all the weights from a folder except the best weigtht"""
    best_weights = None
    if os.path.exists(w_path):
        # remove all but best weight
        all_weights = os.listdir(w_path)
        best_weights = find_best_weight(w_path, best=best, ext=ext)
        ws_to_del = [w for w in all_weights if w != best_weights]
        for w in ws_to_del:
            os.remove(os.path.join(w_path, w))

    return best_weights


def clear_weights(opt_dir, results: dict = None, keep=3, rename=True, write=True):
    """Optimization will save weights of all the trained models, not all of them
    are useful. Here removing weights of all except top 3. The number of models
    whose weights to be retained can be set by `keep` para.
    """
    fname = 'sorted.json'
    if results is None:
        results = make_hpo_results(opt_dir)
        fname = 'sorted_folders.json'

    od = OrderedDict(sorted(results.items()))

    idx = 0
    best_results = {}

    for v in od.values():
        if 'folder' in v:
            folder = v['folder']
            _path = os.path.join(opt_dir, folder)
            w_path = os.path.join(_path, 'weights')

            if idx > keep-1:
                if os.path.exists(w_path):
                    rmtree(w_path)
            else:
                best_weights = remove_all_but_best_weights(w_path)
                best_results[folder] = {'path': _path, 'weights': best_weights}

            idx += 1

    if rename:
        # append ranking of models to folder_names
        idx = 0
        for v in od.values():
            if 'folder' in v:
                folder = v['folder']
                old_path = os.path.join(opt_dir, folder)
                new_path = os.path.join(opt_dir, str(idx+1) + "_" + folder)
                os.rename(old_path, new_path)

                if folder in best_results:
                    best_results[folder] = {'path': new_path, 'weights': best_results[folder]}

                idx += 1

    od = {k: Jsonize(v)() for k, v in od.items()}

    if write:
        sorted_fname = os.path.join(opt_dir, fname)
        with open(sorted_fname, 'w') as sfp:
            json.dump(od, sfp, sort_keys=True, indent=True)

    return best_results


class TrainTestSplit(object):
    """train_test_split of sklearn can not be used for list of arrays so here
    we go
    """
    def __init__(
            self,
            test_fraction: float = 0.3,
            seed : int = None,
            train_indices: Union[list, np.ndarray] = None,
            test_indices: Union[list, np.ndarray] = None
    ):
        """
            test_fraction:
                test fraction. Must be greater than 0. and less than 1.
            seed:
                random seed for reproducibility
        """
        self.test_fraction = test_fraction
        self.state = np.random.RandomState(seed=seed)
        self.train_indices = train_indices
        self.test_indices = test_indices

    def split_by_slicing(
            self,
            x: Union[list, np.ndarray, pd.Series, pd.DataFrame, List[np.ndarray]],
            y: Union[list, np.ndarray, pd.Series, pd.DataFrame, List[np.ndarray]],
    ):
        """splits the x and y by slicing which is defined by `test_fraction`
        Arguments:
            x:
                arrays to split

                - array like such as list, numpy array or pandas dataframe/series
                - list of array like objects
            y:
                array like

                - array like such as list, numpy array or pandas dataframe/series
                - list of array like objects
        """
        def split_arrays(array):

            if isinstance(array, list):
                # x is list of arrays
                # assert that all arrays are of equal length
                assert len(set([len(_array) for _array in array])) == 1, f"arrays are of not same length"
                split_at = int(array[0].shape[0] * (1. - self.test_fraction))
            else:
                split_at = int(len(array) * (1. - self.test_fraction))

            train, test = (self.slice_arrays(array, 0, split_at), self.slice_arrays(array, split_at))

            return train, test

        train_x, test_x = split_arrays(x)
        train_y, test_y = split_arrays(y)

        return train_x, test_x, train_y, test_y

    def split_by_random(
            self,
            x: Union[list, np.ndarray, pd.Series, pd.DataFrame, List[np.ndarray]],
            y: Union[list, np.ndarray, pd.Series, pd.DataFrame, List[np.ndarray]],
    ):
        """
        splits the x and y by random splitting.
        Arguments:
            x:
                arrays to split

                - array like such as list, numpy array or pandas dataframe/series
                - list of array like objects
            y:
                array like

                - array like such as list, numpy array or pandas dataframe/series
                - list of array like objects

        """
        if isinstance(x, list):
            indices = np.arange(len(x[0]))
        else:
            indices = np.arange(len(x))

        indices = self.state.permutation(indices)

        split_at = int(len(indices) * (1. - self.test_fraction))
        train_indices, test_indices = (self.slice_arrays(indices, 0, split_at), self.slice_arrays(indices, split_at))

        train_x = self.slice_with_indices(x, train_indices)
        train_y = self.slice_with_indices(y, train_indices)

        test_x = self.slice_with_indices(x, test_indices)
        test_y = self.slice_with_indices(y, test_indices)

        return train_x, test_x, train_y, test_y

    def split_by_indices(
            self,
            x: Union[list, np.ndarray, pd.Series, pd.DataFrame, List[np.ndarray]],
            y: Union[list, np.ndarray, pd.Series, pd.DataFrame, List[np.ndarray]],
    ):
        """splits the x and y by user defined `train_indices` and `test_indices`"""

        return self.slice_with_indices(x, self.train_indices), \
               self.slice_with_indices(x, self.test_indices), \
               self.slice_with_indices(y, self.train_indices), \
               self.slice_with_indices(y, self.test_indices)

    @staticmethod
    def slice_with_indices(array, indices):
        if isinstance(array, list):
            _data = []

            for d in array:
                assert isinstance(d, np.ndarray)
                _data.append(d[indices])
        else:
            assert isinstance(array, np.ndarray) or isinstance(array, pd.DatetimeIndex)
            _data = array[indices]
        return _data

    @staticmethod
    def slice_arrays(arrays, start, stop=None):
        if isinstance(arrays, list):
            return [array[start:stop] for array in arrays]
        elif hasattr(arrays, 'shape'):
            return arrays[start:stop]

    def KFold_splits(
            self,
            x,
            y,
            n_splits,
            shuffle=True,
            random_state=None
    ):
        from sklearn.model_selection import KFold
        kf = KFold(n_splits=n_splits, random_state=random_state,  shuffle=shuffle)
        spliter = kf.split(x[0] if isinstance(x, list) else x)

        for tr_idx, test_idx in spliter:

            if isinstance(x, list):
                train_x = [xarray[tr_idx] for xarray in x]
                test_x = [xarray[test_idx] for xarray in x]
            else:
                train_x = x[tr_idx]
                test_x = x[test_idx]

            if isinstance(y, list):
                train_y = [yarray[tr_idx] for yarray in y]
                test_y = [yarray[test_idx] for yarray in y]
            else:
                train_y = y[tr_idx]
                test_y = y[test_idx]

            yield (train_x, train_y), (test_x, test_y)


def ts_features(data: Union[np.ndarray, pd.DataFrame, pd.Series],
                precision: int = 3,
                name: str = '',
                st: int = 0,
                en: int = None,
                features: Union[list, str] = None
                ) -> dict:
    """
    Extracts features from 1d time series data. Features can be
        * point, one integer or float point value for example mean
        * 1D, 1D array for example sin(data)
        * 2D, 2D array for example wavelent transform
    Arguments:
        Gets all the possible stats about an array like object `data`.
        data: array like
        precision: number of significant figures
        name: str, only for erro or warning messages
        st: str/int, starting index of data to be considered.
        en: str/int, end index of data to be considered.
        features: name/names of features to extract from data.
    # information holding degree
    """
    point_features = {
        'Skew': skew,
        'Kurtosis': kurtosis,
        'Mean': np.nanmean,
        'Geometric Mean': gmean,
        'Standard error of mean': scipy.stats.sem,
        'Median': np.nanmedian,
        'Variance': np.nanvar,
        'Coefficient of Variation': variation,
        'Std': np.nanstd,
        'Non Zeros': np.count_nonzero,
        'Min': np.nanmin,
        'Max': np.nanmax,
        'Sum': np.nansum,
        'Counts': np.size
    }

    point_features_lambda = {
        'Shannon entropy': lambda x: np.round(scipy.stats.entropy(pd.Series(x).value_counts()), precision),
        'Negative counts': lambda x: int(np.sum(x < 0.0)),
        '90th percentile': lambda x: np.round(np.nanpercentile(x, 90), precision),
        '75th percentile': lambda x: np.round(np.nanpercentile(x, 75), precision),
        '50th percentile': lambda x: np.round(np.nanpercentile(x, 50), precision),
        '25th percentile': lambda x: np.round(np.nanpercentile(x, 25), precision),
        '10th percentile': lambda x: np.round(np.nanpercentile(x, 10), precision),
    }

    if not isinstance(data, np.ndarray):
        if hasattr(data, '__len__'):
            data = np.array(data)
        else:
            raise TypeError(f"{name} must be array like but it is of type {data.__class__.__name__}")

    if np.array(data).dtype.type is np.str_:
        warnings.warn(f"{name} contains string values")
        return {}

    if 'int' not in data.dtype.name:
        if 'float' not in data.dtype.name:
            warnings.warn(f"changing the dtype of {name} from {data.dtype.name} to float")
            data = data.astype(np.float64)

    assert data.size == len(data), f"""
data must be 1 dimensional array but it has shape {np.shape(data)}
"""
    data = data[st:en]
    stats = dict()

    if features is None:
        features = list(point_features.keys()) + list(point_features_lambda.keys())
    elif isinstance(features, str):
        features = [features]

    for feat in features:
        if feat in point_features:
            stats[feat] = np.round(point_features[feat](data), precision)
        elif feat in point_features_lambda:
            stats[feat] = point_features_lambda[feat](data)

    if 'Harmonic Mean' in features:
        try:
            stats['Harmonic Mean'] = np.round(hmean(data), precision)
        except ValueError:
            warnings.warn(f"""Unable to calculate Harmonic mean for {name}. Harmonic mean only defined if all
                          elements are greater than or equal to zero""", UserWarning)

    return Jsonize(stats)()


def prepare_data(
        data: np.ndarray,
        lookback: int,
        num_inputs: int = None,
        num_outputs: int = None,
        input_steps: int = 1,
        forecast_step: int = 0,
        forecast_len: int = 1,
        known_future_inputs: bool = False,
        output_steps: int = 1,
        mask: Union[int, float, np.ndarray] = None
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    converts a numpy nd array into a supervised machine learning problem.

    Parameters
    ----------
        data :
            nd numpy array whose first dimension represents the number
            of examples and the second dimension represents the number of features.
            Some of those features will be used as inputs and some will be considered
            as outputs depending upon the values of `num_inputs` and `num_outputs`.
        lookback :
            number of previous steps/values to be used at one step.
        num_inputs :
            default None, number of input features in data. If None,
            it will be calculated as features-outputs. The input data will be all
            from start till num_outputs in second dimension.
        num_outputs :
            number of columns (from last) in data to be used as output.
            If None, it will be caculated as features-inputs.
        input_steps:
            strides/number of steps in input data
        forecast_step :
            must be greater than equal to 0, which t+ith value to
            use as target where i is the horizon. For time series prediction, we
            can say, which horizon to predict.
        forecast_len :
            number of horizons/future values to predict.
        known_future_inputs :
            Only useful if `forecast_len`>1. If True, this
            means, we know and use 'future inputs' while making predictions at t>0
        output_steps :
            step size in outputs. If =2, it means we want to predict
            every second value from the targets
        mask :
            If int, then the examples with these values in
            the output will be skipped. If array then it must be a boolean mask
            indicating which examples to include/exclude. The length of mask should
            be equal to the number of generated examples. The number of generated
            examples is difficult to prognose because it depend upon lookback, input_steps,
            and forecast_step. Thus it is better to provide an integer indicating
            which values in outputs are to be considered as invalid. Default is
            None, which indicates all the generated examples will be returned.

    Returns
    -------
        x : numpy array of shape (examples, lookback, ins) consisting of
        input examples
        prev_y : numpy array consisting of previous outputs
        y : numpy array consisting of target values

    Given following data consisting of input/output pairs

    +--------+--------+---------+---------+----------+
    | input1 | input2 | output1 | output2 | output 3 |
    +========+========+=========+=========+==========+
    |   1    |   11   |   21    |    31   |   41     |
    +--------+--------+---------+---------+----------+
    |   2    |   12   |   22    |    32   |   42     |
    +--------+--------+---------+---------+----------+
    |   3    |   13   |   23    |    33   |   43     |
    +--------+--------+---------+---------+----------+
    |   4    |   14   |   24    |    34   |   44     |
    +--------+--------+---------+---------+----------+
    |   5    |   15   |   25    |    35   |   45     |
    +--------+--------+---------+---------+----------+
    |   6    |   16   |   26    |    36   |   46     |
    +--------+--------+---------+---------+----------+
    |   7    |   17   |   27    |    37   |   47     |
    +--------+--------+---------+---------+----------+

    If we use following 2 time series as input

    +--------+--------+
    | input1 | input2 |
    +========+========+
    |  1     |  11    |
    +--------+--------+
    |     2  |  12    |
    +--------+--------+
    | 3      |  13    |
    +--------+--------+
    | 4      |  14    |
    +--------+--------+
    | 5      |  15    |
    +--------+--------+
    | 6      |  16    |
    +--------+--------+
    | 7      |  17    |
    +--------+--------+

    then  ``num_inputs``=2, ``lookback``=7, ``input_steps``=1

    and if we want to predict

    +---------+---------+----------+
    | output1 | output2 | output 3 |
    +=========+=========+==========+
    |   27    |   37    |   47     |
    +---------+---------+----------+

    then ``num_outputs``=3, ``forecast_len``=1,  ``forecast_step``=0,

    if we want to predict

    +---------+---------+----------+
    | output1 | output2 | output 3 |
    +=========+=========+==========+
    | 28      | 38      | 48       |
    +---------+---------+----------+

    then ``num_outputs``=3, ``forecast_len``=1,  ``forecast_step``=1,

    if we want to predict

    +---------+---------+----------+
    | output1 | output2 | output 3 |
    +=========+=========+==========+
    |  27     |  37     |  47      |
    +---------+---------+----------+
    |  28     |  38     |  48      |
    +---------+---------+----------+

    then ``num_outputs``=3, forecast_len=2,  horizon/forecast_step=0,

    if we want to predict

    +---------+---------+----------+
    | output1 | output2 | output 3 |
    +=========+=========+==========+
    |   28    |   38    |   48     |
    +---------+---------+----------+
    |   29    |   39    |   49     |
    +---------+---------+----------+
    |   30    |   40    |   50     |
    +---------+---------+----------+

    then ``num_outputs`` =3, ``forecast_len`` =3,  ``forecast_step`` =1,

    if we want to predict

    +---------+
    | output2 |
    +=========+
    |   38    |
    +---------+
    |   39    |
    +---------+
    |   40    |

    then ``num_outputs``=1, ``forecast_len``=3, ``forecast_step``=0

    if we predict

    +---------+
    | output2 |
    +=========+
    | 39      |
    +---------+

    then ``num_outputs``=1, ``forecast_len``=1, ``forecast_step``=2

    if we predict

    +---------+
    | output2 |
    +=========+
    | 39      |
    +---------+
    | 40      |
    +---------+
    | 41      |
    +---------+

     then ``num_outputs``=1, ``forecast_len``=3, ``forecast_step``=2

    If we use following two time series as input

    +--------+--------+
    |input1  | input2 |
    +========+========+
    |   1    |  11    |
    +--------+--------+
    |   3    |  13    |
    +--------+--------+
    |   5    |  15    |
    +--------+--------+
    |   7    |  17    |
    +--------+--------+

    then   ``num_inputs``=2, ``lookback``=4, ``input_steps``=2

    If the input is

    +--------+--------+
    | input1 | input2 |
    +========+========+
    |    1   |  11    |
    +--------+--------+
    |    2   |  12    |
    +--------+--------+
    |    3   |  13    |
    +--------+--------+
    |    4   |  14    |
    +--------+--------+
    |    5   |  15    |
    +--------+--------+
    |    6   |  16    |
    +--------+--------+
    |   7    |  17    |
    +--------+--------+

    and target/output is

    +---------+---------+----------+
    | output1 | output2 | output 3 |
    +=========+=========+==========+
    |    25   |    35   |    45    |
    +---------+---------+----------+
    |    26   |    36   |    46    |
    +---------+---------+----------+
    |    27   |    37   |    47    |
    +---------+---------+----------+

    This means we make use of 'known future inputs'. This can be achieved using following configuration
    num_inputs=2, num_outputs=3, lookback=4, forecast_len=3, forecast_step=1, known_future_inputs=True

    The general shape of output/target/label is
    (examples, num_outputs, forecast_len)

    The general shape of inputs/x is
    (examples, lookback + forecast_len-1, ....num_inputs)


    Examples:
        >>> import numpy as np
        >>> from ai4water.utils.utils import prepare_data
        >>> num_examples = 50
        >>> dataframe = np.arange(int(num_examples*5)).reshape(-1, num_examples).transpose()
        >>> dataframe[0:10]
        array([[  0,  50, 100, 150, 200],
               [  1,  51, 101, 151, 201],
               [  2,  52, 102, 152, 202],
               [  3,  53, 103, 153, 203],
               [  4,  54, 104, 154, 204],
               [  5,  55, 105, 155, 205],
               [  6,  56, 106, 156, 206],
               [  7,  57, 107, 157, 207],
               [  8,  58, 108, 158, 208],
               [  9,  59, 109, 159, 209]])
        >>> x, prevy, y = prepare_data(data, num_outputs=2, lookback=4,
        ...    input_steps=2, forecast_step=2, forecast_len=4)
        >>> x[0]
        array([[  0.,  50., 100.],
              [  2.,  52., 102.],
              [  4.,  54., 104.],
              [  6.,  56., 106.]], dtype=float32)
        >>> y[0]
        array([[158., 159., 160., 161.],
              [208., 209., 210., 211.]], dtype=float32)

        >>> x, prevy, y = prepare_data(data, num_outputs=2, lookback=4,
        ...    forecast_len=3, known_future_inputs=True)
        >>> x[0]
        array([[  0,  50, 100],
               [  1,  51, 101],
               [  2,  52, 102],
               [  3,  53, 103],
               [  4,  54, 104],
               [  5,  55, 105],
               [  6,  56, 106]])       # (7, 3)
        >>> # it is important to note that although lookback=4 but x[0] has shape of 7
        >>> y[0]
        array([[154., 155., 156.],
               [204., 205., 206.]], dtype=float32)  # (2, 3)
    """
    if not isinstance(data, np.ndarray):
        if isinstance(data, pd.DataFrame):
            data = data.values
        else:
            raise TypeError(f"unknown data type for data {data.__class__.__name__}")

    if num_inputs is None and num_outputs is None:
        raise ValueError("""
Either of num_inputs or num_outputs must be provided.
""")

    features = data.shape[1]
    if num_outputs is None:
        num_outputs = features - num_inputs

    if num_inputs is None:
        num_inputs = features - num_outputs

    assert num_inputs + num_outputs == features, f"""
num_inputs {num_inputs} + num_outputs {num_outputs} != total features {features}"""

    if len(data) <= 1:
        raise ValueError(f"Can not create batches from data with shape {data.shape}")

    time_steps = lookback
    if known_future_inputs:
        lookback = lookback + forecast_len
        assert forecast_len > 1, f"""
            known_futre_inputs should be True only when making predictions at multiple 
            horizons i.e. when forecast length/number of horizons to predict is > 1.
            known_future_inputs: {known_future_inputs}
            forecast_len: {forecast_len}"""

    examples = len(data)

    x = []
    prev_y = []
    y = []

    for i in range(examples - lookback * input_steps + 1 - forecast_step - forecast_len + 1):
        stx, enx = i, i + lookback * input_steps
        x_example = data[stx:enx:input_steps, 0:features - num_outputs]

        st, en = i, i + (lookback - 1) * input_steps
        y_data = data[st:en:input_steps, features - num_outputs:]

        sty = (i + time_steps * input_steps) + forecast_step - input_steps
        eny = sty + forecast_len
        target = data[sty:eny, features - num_outputs:]

        x.append(np.array(x_example))
        prev_y.append(np.array(y_data))
        y.append(np.array(target))

    x = np.stack(x)
    prev_y = np.array([np.array(i, dtype=np.float32) for i in prev_y], dtype=np.float32)
    # transpose because we want labels to be of shape (examples, outs, forecast_len)
    y = np.array([np.array(i, dtype=np.float32).T for i in y], dtype=np.float32)

    if mask is not None:
        if isinstance(mask, np.ndarray):
            assert mask.ndim == 1
            assert len(x) == len(mask), f"Number of generated examples are {len(x)} " \
                                        f"but the length of mask is {len(mask)}"
        elif isinstance(mask, float) and np.isnan(mask):
            mask = np.invert(np.isnan(y))
            mask = np.array([all(i.reshape(-1,)) for i in mask])
        else:
            assert isinstance(mask, int), f"""
                    Invalid mask identifier given of type: {mask.__class__.__name__}"""
            mask = y != mask
            mask = np.array([all(i.reshape(-1,)) for i in mask])

        x = x[mask]
        prev_y = prev_y[mask]
        y = y[mask]

    return x, prev_y, y


def find_tot_plots(features, max_subplots):

    tot_plots = np.linspace(0, features, int(features / max_subplots) + 1 if features % max_subplots == 0 else int(
        features / max_subplots) + 2)
    # converting each value to int because linspace can return array containing floats if features is odd
    tot_plots = [int(i) for i in tot_plots]
    return tot_plots



class JsonEncoder(json.JSONEncoder):

    def default(self, obj):
        if 'int' in obj.__class__.__name__:
            return int(obj)
        elif 'float' in obj.__class__.__name__:
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif 'bool' in obj.__class__.__name__:
            return bool(obj)
        elif callable(obj) and hasattr(obj, '__module__'):
            return obj.__module__
        else:
            return super(JsonEncoder, self).default(obj)


def plot_activations_along_inputs(
        data: np.ndarray,
        activations: np.ndarray,
        observations: np.ndarray,
        predictions: np.ndarray,
        in_cols: list,
        out_cols: list,
        lookback: int,
        name: str,
        path: str,
        vmin=None,
        vmax=None,
        show=False
):

    # activation must be of shape (num_examples, lookback, input_features)
    assert activations.shape[1] == lookback
    assert activations.shape[2] == len(in_cols), f'{activations.shape}, {len(in_cols)}'

    # data is of shape (num_examples, input_features)
    assert data.shape[1] == len(in_cols)

    assert len(data) == len(activations)

    for out in range(len(out_cols)):
        pred = predictions[:, out]
        obs = observations[:, out]
        out_name = out_cols[out]

        for idx in range(len(in_cols)):
            plt.close('all')
            fig, (ax1, ax2, ax3) = plt.subplots(3, sharex='all')
            fig.set_figheight(12)

            ax1.plot(data[:, idx], label=in_cols[idx])
            ax1.legend()
            ax1.set_title('activations w.r.t ' + in_cols[idx])
            ax1.set_ylabel(in_cols[idx])

            ax2.plot(pred, label='Prediction')
            ax2.plot(obs, '.', label='Observed')
            ax2.legend()
            ytick_labels = [f"t-{int(i)}" for i in np.linspace(lookback - 1, 0, lookback)]
            axis, im = imshow(activations[:, :, idx].transpose(),
                              vmin=vmin,
                              vmax=vmax,
                              aspect="auto",
                              ax = ax3,
                              xlabel="Examples",
                              ylabel="lookback steps",
                              show=False,
                              yticklabels=ytick_labels)
            fig.colorbar(im, orientation='horizontal', pad=0.2)
            plt.subplots_adjust(wspace=0.005, hspace=0.005)
            _name = f'attn_weights_{out_name}_{name}_'
            plt.savefig(os.path.join(path, _name) + in_cols[idx], dpi=400, bbox_inches='tight')

            if show:
                plt.show()

            plt.close('all')

    return


def print_something(something, prefix=''):
    """prints shape of some python object"""
    if isinstance(something, np.ndarray):
        print(f"{prefix} shape: ", something.shape)
    elif isinstance(something, list):
        print(f"{prefix} shape: ", [thing.shape for thing in something if isinstance(thing, np.ndarray)])
    elif isinstance(something, dict):
        print(f"{prefix} shape: ")
        pprint.pprint({k: v.shape for k, v in something.items()}, width=40)
    elif something is not None:
        print(f"{prefix} shape: ", something.shape)
        print(something)
    else:
        print(something)


def maybe_three_outputs(data, teacher_forcing=False):
    """num_outputs: how many outputs from data we want"""
    if teacher_forcing:
        num_outputs = 3
    else:
        num_outputs = 2

    if num_outputs == 2:
        if len(data) == 2:
            return data[0], data[1]
        elif len(data) == 3:
            return data[0], data[2]
    else:
        if len(data)==3:
            return [data[0], data[1]], data[2]
        # DA, IA-LSTM models return [x,prevy],y even when teacher_forcing is on!
        return data


def get_version_info(
        **kwargs
) -> dict:
    # todo, chekc which attributes are not available in different versions
    import sys
    info = {'python': sys.version, 'os': os.name}
    if kwargs.get('tf', None):
        tf = kwargs['tf']
        info['tf_is_built_with_cuda'] = tf.test.is_built_with_cuda()
        info['is_built_with_gpu_support'] = tf.test.is_built_with_gpu_support()
        info['tf_is_gpu_available'] = tf.test.is_gpu_available()
        info['eager_execution'] = tf.executing_eagerly()

    for k, v in kwargs.items():
        if v is not None:
            info[k] = getattr(v, '__version__', 'NotDefined')

    return info

def check_attributes(model, attributes):
    for method in attributes:
        if not hasattr(model, method):
            raise ValueError(f"your custom class does not have {method}")


def get_nrows_ncols(n_rows, n_subplots)->"tuple[int, int]":

    if n_rows is None:
        n_rows = int(np.sqrt(n_subplots))
    n_cols = max(int(n_subplots / n_rows), 1)  # ensure n_cols != 0
    n_rows = int(n_subplots / n_cols)

    while not ((n_subplots / n_cols).is_integer() and
               (n_subplots / n_rows).is_integer()):
        n_cols -= 1
        n_rows = int(n_subplots / n_cols)
    return n_rows, n_cols



from typing import Union

import os
import pandas as pd

from .camels import Camels
from .camels import CAMELS_AUS
from .camels import CAMELS_CL
from .camels import CAMELS_BR
from .camels import CAMELS_GB
from .camels import CAMELS_US
from .camels import LamaH
from .camels import HYSETS
from .camels import HYPE

from .datasets import Weisssee
from .datasets import WaterChemEcuador
from .datasets import WaterChemVictoriaLakes
from .datasets import WeatherJena
from .datasets import WQCantareira
from .datasets import WQJordan
from .datasets import FlowSamoylov
from .datasets import FlowSedDenmark
from .datasets import StreamTempSpain
from .datasets import RiverTempEroo
from .datasets import HoloceneTemp
from .datasets import FlowTetRiver
from .datasets import SedimentAmersee
from .datasets import HydrocarbonsGabes
from .datasets import HydroChemJava
from .datasets import PrecipBerlin
from .datasets import GeoChemMatane
from .datasets import WQJordan2
from .datasets import YamaguchiClimateJp
from .datasets import FlowBenin
from .datasets import HydrometricParana
from .datasets import RiverTempSpain
from .datasets import RiverIsotope
from .datasets import EtpPcpSamoylov
from .datasets import SWECanada
from .datasets import MtropicsLaos
from .datasets import MtropcsThailand
from .datasets import MtropicsVietnam


def busan_beach(inputs: list = None,
                target: Union[list, str] = 'tetx_coppml'
                ) -> pd.DataFrame:
    """
    Loads the Antibiotic resitance genes (ARG) data from a recreational beach
    in Busan, South Korea along with environment variables.

    The data is in the form of
    mutlivariate time series and was collected over the period of 2 years during
    several precipitation events. The frequency of environmental data is 30 mins
    while the ARG is discontinuous. The data and its pre-processing is described
    in detail in Jang_ et al., 2021

    Arguments
    ---------
        inputs :
            features to use as input. By default all environmental data
            is used which consists of following parameters

            - tide_cm
            - wat_temp_c
            - sal_psu
            - air_temp_c
            - pcp_mm
            - pcp3_mm
            - pcp6_mm
            - pcp12_mm
            - wind_dir_deg
            - wind_speed_mps
            - air_p_hpa
            - mslp_hpa
            - rel_hum

        target :
            feature/features to use as target/output. By default
            `tetx_coppml` is used as target.
            Logically one or more from following can be considered as target

            - ecoli
            - 16s
            - inti1
            - Total_args
            - tetx_coppml
            - sul1_coppml
            - blaTEM_coppml
            - aac_coppml
            - Total_otus
            - otu_5575
            - otu_273
            - otu_94

    Returns
    -------
        a pandas dataframe with inputs and target and indexed
            with pandas.DateTimeIndex

    Example:
        >>> from ai4water.datasets import busan_beach
        >>> dataframe = busan_beach()

    .. _Jang:
        https://doi.org/10.1016/j.watres.2021.117001
    """
    fpath = os.path.join(os.path.dirname(__file__), "arg_busan.csv")
    df = pd.read_csv(fpath, index_col="index")
    df.index = pd.to_datetime(df.index)

    default_inputs = ['tide_cm', 'wat_temp_c', 'sal_psu', 'air_temp_c', 'pcp_mm', 'pcp3_mm', 'pcp6_mm',
                      'pcp12_mm', 'wind_dir_deg', 'wind_speed_mps', 'air_p_hpa', 'mslp_hpa', 'rel_hum'
                      ]
    default_targets = [col for col in df.columns if col not in default_inputs]

    if inputs is None:
        inputs = default_inputs

    if not isinstance(target, list):
        if isinstance(target, str):
            target = [target]
    elif isinstance(target, list):
        pass
    else:
        target = default_targets

    assert isinstance(target, list)

    df = df[inputs + target]

    return df


def load_nasdaq(inputs: Union[str, list, None] = None, target: str = 'NDX'):
    """Loads Nasdaq100 by downloading it if it is not already downloaded."""
    fname = os.path.join(os.path.dirname(__file__), "nasdaq100_padding.csv")

    if not os.path.exists(fname):
        print(f"downloading file to {fname}")
        df = pd.read_csv("https://raw.githubusercontent.com/KurochkinAlexey/DA-RNN/master/nasdaq100_padding.csv")
        df.to_csv(fname)

    df = pd.read_csv(fname)
    in_cols = list(df.columns)
    in_cols.remove(target)
    if inputs is None:
        inputs = in_cols
    target = [target]

    return df[inputs + target]


all_datasets = ['CAMELS_AUS', 'CAMELS_CL', 'CAMELS_US', 'CAMELS_GB', 'CAMELS_BR',
                'CAMELS_CL', 'LamaH', 'HYPE',
                'HYSETS']
